@@ -137,21 +137,29 @@ class CommonIdentityTests(object):
         """The API version being tested."""
 
     def test_discovering(self):
+        disc = fixture.DiscoveryList(v2=False, v3=False)
+
+        disc.add_nova_microversion(
+            href=self.TEST_COMPUTE_ADMIN,
+            id='v2.1', status='CURRENT',
+            min_version='2.1', version='2.38')
+
         self.stub_url('GET', [],
                       base_url=self.TEST_COMPUTE_ADMIN,
-                      json=self.TEST_DISCOVERY)
+                      json=disc)
 
         body = 'SUCCESS'
 
         # which gives our sample values
-        self.stub_url('GET', ['path'], text=body)
+        self.stub_url('GET', ['path'],
+                      text=body, base_url=self.TEST_COMPUTE_ADMIN)
 
         a = self.create_auth_plugin()
         s = session.Session(auth=a)
 
         resp = s.get('/path', endpoint_filter={'service_type': 'compute',
                                                'interface': 'admin',
-                                               'version': self.version})
+                                               'version': '2.1'})
 
         self.assertEqual(200, resp.status_code)
         self.assertEqual(body, resp.text)
@@ -169,13 +177,21 @@ class CommonIdentityTests(object):
         self.assertEqual(new_body, resp.text)
 
     def test_discovery_uses_provided_session_cache(self):
+        disc = fixture.DiscoveryList(v2=False, v3=False)
+
+        disc.add_nova_microversion(
+            href=self.TEST_COMPUTE_ADMIN,
+            id='v2.1', status='CURRENT',
+            min_version='2.1', version='2.38')
+
         # register responses such that if the discovery URL is hit more than
         # once then the response will be invalid and not point to COMPUTE_ADMIN
-        resps = [{'json': self.TEST_DISCOVERY}, {'status_code': 500}]
+        resps = [{'json': disc}, {'status_code': 500}]
         self.requests_mock.get(self.TEST_COMPUTE_ADMIN, resps)
 
         body = 'SUCCESS'
-        self.stub_url('GET', ['path'], text=body)
+        self.stub_url('GET', ['path'],
+                      text=body, base_url=self.TEST_COMPUTE_ADMIN)
 
         cache = {}
         # now either of the two plugins I use, it should not cause a second
@@ -189,23 +205,32 @@ class CommonIdentityTests(object):
                          auth=auth,
                          endpoint_filter={'service_type': 'compute',
                                           'interface': 'admin',
-                                          'version': self.version})
+                                          'version': '2.1'})
 
             self.assertEqual(200, resp.status_code)
             self.assertEqual(body, resp.text)
         self.assertIn(self.TEST_COMPUTE_ADMIN, cache.keys())
 
     def test_discovery_uses_session_cache(self):
+        disc = fixture.DiscoveryList(v2=False, v3=False)
+
+        disc.add_nova_microversion(
+            href=self.TEST_COMPUTE_ADMIN,
+            id='v2.1', status='CURRENT',
+            min_version='2.1', version='2.38')
+
         # register responses such that if the discovery URL is hit more than
         # once then the response will be invalid and not point to COMPUTE_ADMIN
-        resps = [{'json': self.TEST_DISCOVERY}, {'status_code': 500}]
+        resps = [{'json': disc}, {'status_code': 500}]
         self.requests_mock.get(self.TEST_COMPUTE_ADMIN, resps)
 
         body = 'SUCCESS'
-        self.stub_url('GET', ['path'], text=body)
+        self.stub_url('GET', ['path'],
+                      base_url=self.TEST_COMPUTE_ADMIN,
+                      text=body)
 
         filter = {'service_type': 'compute', 'interface': 'admin',
-                  'version': self.version}
+                  'version': '2.1'}
 
         # create a session and call the endpoint, causing its cache to be set
         sess = session.Session()
@@ -224,13 +249,22 @@ class CommonIdentityTests(object):
             self.assertEqual(body, resp.text)
 
     def test_discovery_uses_plugin_cache(self):
+        disc = fixture.DiscoveryList(v2=False, v3=False)
+
+        disc.add_nova_microversion(
+            href=self.TEST_COMPUTE_ADMIN,
+            id='v2.1', status='CURRENT',
+            min_version='2.1', version='2.38')
+
         # register responses such that if the discovery URL is hit more than
         # once then the response will be invalid and not point to COMPUTE_ADMIN
-        resps = [{'json': self.TEST_DISCOVERY}, {'status_code': 500}]
+        resps = [{'json': disc}, {'status_code': 500}]
         self.requests_mock.get(self.TEST_COMPUTE_ADMIN, resps)
 
         body = 'SUCCESS'
-        self.stub_url('GET', ['path'], text=body)
+        self.stub_url('GET', ['path'],
+                      base_url=self.TEST_COMPUTE_ADMIN,
+                      text=body)
 
         # now either of the two sessions I use, it should not cause a second
         # request to the discovery url. Calling discovery directly should also
@@ -244,22 +278,31 @@ class CommonIdentityTests(object):
                             auth=auth,
                             endpoint_filter={'service_type': 'compute',
                                              'interface': 'admin',
-                                             'version': self.version})
+                                             'version': '2.1'})
 
             self.assertEqual(200, resp.status_code)
             self.assertEqual(body, resp.text)
 
     def test_discovery_uses_session_plugin_cache(self):
+        disc = fixture.DiscoveryList(v2=False, v3=False)
+
+        disc.add_nova_microversion(
+            href=self.TEST_COMPUTE_ADMIN,
+            id='v2.1', status='CURRENT',
+            min_version='2.1', version='2.38')
+
         # register responses such that if the discovery URL is hit more than
         # once then the response will be invalid and not point to COMPUTE_ADMIN
-        resps = [{'json': self.TEST_DISCOVERY}, {'status_code': 500}]
+        resps = [{'json': disc}, {'status_code': 500}]
         self.requests_mock.get(self.TEST_COMPUTE_ADMIN, resps)
 
         body = 'SUCCESS'
-        self.stub_url('GET', ['path'], text=body)
+        self.stub_url('GET', ['path'],
+                      base_url=self.TEST_COMPUTE_ADMIN,
+                      text=body)
 
         filter = {'service_type': 'compute', 'interface': 'admin',
-                  'version': self.version}
+                  'version': '2.1'}
 
         # create a plugin and call the endpoint, causing its cache to be set
         plugin = self.create_auth_plugin()
@@ -286,7 +329,7 @@ class CommonIdentityTests(object):
         sb = session.Session()
         discovery_cache = {}
 
-        expected_url = urllib.parse.urljoin(self.TEST_ROOT_URL, '/v2.0')
+        expected_url = urllib.parse.urljoin(self.TEST_COMPUTE_ADMIN, '/v2.0')
         for sess in (sa, sb):
 
             disc = discover.get_discovery(
@@ -385,6 +428,40 @@ class CommonIdentityTests(object):
         self.assertEqual(v2_compute, url_v2)
         self.assertEqual(v3_compute, url_v3)
 
+    def test_discovering_version_no_discovery(self):
+
+        a = self.create_auth_plugin()
+        s = session.Session(auth=a)
+
+        # Grab a version that can be returned without doing discovery
+        # This tests that it doesn't make a discovery call because we don't
+        # have a reqquest mock, and this will throw an exception if it tries
+        version = s.get_api_major_version(
+            service_type='volumev2', interface='admin')
+        self.assertEqual((2, 0), version)
+
+    def test_discovering_version_with_discovery(self):
+
+        a = self.create_auth_plugin()
+        s = session.Session(auth=a)
+
+        v2_compute = self.TEST_COMPUTE_ADMIN + '/v2.0'
+        v3_compute = self.TEST_COMPUTE_ADMIN + '/v3'
+
+        disc = fixture.DiscoveryList(v2=False, v3=False)
+        disc.add_v2(v2_compute)
+        disc.add_v3(v3_compute)
+
+        self.stub_url('GET', [], base_url=self.TEST_COMPUTE_ADMIN, json=disc)
+
+        # This needs to do version discovery to find the version
+        version = s.get_api_major_version(
+            service_type='compute', interface='admin')
+        self.assertEqual((3, 0), version)
+        self.assertEqual(
+            self.requests_mock.request_history[-1].url,
+            self.TEST_COMPUTE_ADMIN)
+
     def test_direct_discovering_with_relative_link(self):
         # need to construct list this way for relative
         disc = fixture.DiscoveryList(v2=False, v3=False)
@@ -499,6 +576,8 @@ class CommonIdentityTests(object):
                                    interface='admin')
 
         self.assertEqual(self.TEST_COMPUTE_ADMIN + '/v3', data.url)
+        # We should have gotten the version from the URL
+        self.assertEqual((3, 0), data.api_version)
 
     def test_endpoint_data_no_version_no_discovery(self):
         a = self.create_auth_plugin()
@@ -510,6 +589,22 @@ class CommonIdentityTests(object):
                                    discover_versions=False)
 
         self.assertEqual(self.TEST_COMPUTE_ADMIN, data.url)
+        # There's no version in the URL and no document - we have no idea
+        self.assertIsNone(data.api_version)
+
+    def test_endpoint_data_version_url_no_discovery(self):
+        a = self.create_auth_plugin()
+        s = session.Session(auth=a)
+
+        data = a.get_endpoint_data(session=s,
+                                   service_type='volumev3',
+                                   interface='admin',
+                                   discover_versions=False)
+
+        self.assertEqual(
+            self.TEST_VOLUME.versions['v3'].service.admin, data.url)
+        # There's v3 in the URL
+        self.assertEqual((3, 0), data.api_version)
 
     def test_endpoint_no_version(self):
         a = self.create_auth_plugin()
@@ -589,6 +684,37 @@ class CommonIdentityTests(object):
             self.assertEqual(v3_compute, v3_data.service_url)
             self.assertEqual(self.TEST_COMPUTE_ADMIN, v3_data.catalog_url)
 
+    def test_get_current_versioned_data(self):
+        v2_compute = self.TEST_COMPUTE_ADMIN + '/v2.0'
+        v3_compute = self.TEST_COMPUTE_ADMIN + '/v3'
+
+        disc = fixture.DiscoveryList(v2=False, v3=False)
+        disc.add_v2(v2_compute)
+        disc.add_v3(v3_compute)
+
+        # Make sure that we don't do more than one discovery call
+        # register responses such that if the discovery URL is hit more than
+        # once then the response will be invalid and not point to COMPUTE_ADMIN
+        resps = [{'json': disc}, {'status_code': 500}]
+        self.requests_mock.get(self.TEST_COMPUTE_ADMIN, resps)
+
+        a = self.create_auth_plugin()
+        s = session.Session(auth=a)
+
+        data = a.get_endpoint_data(session=s,
+                                   service_type='compute',
+                                   interface='admin')
+        self.assertEqual(v3_compute, data.url)
+
+        v3_data = data.get_current_versioned_data(s)
+
+        self.assertEqual(v3_compute, v3_data.url)
+        self.assertEqual(v3_compute, v3_data.service_url)
+        self.assertEqual(self.TEST_COMPUTE_ADMIN, v3_data.catalog_url)
+        self.assertEqual((3, 0), v3_data.api_version)
+        self.assertIsNone(v3_data.min_microversion)
+        self.assertIsNone(v3_data.max_microversion)
+
     def test_interface_list(self):
 
         a = self.create_auth_plugin()
@@ -676,8 +802,8 @@ class CommonIdentityTests(object):
         self.assertEqual(self.TEST_VOLUME.versions['v2'].service.public,
                          v2_data.service_url)
         self.assertEqual(self.TEST_VOLUME.catalog.public, v2_data.catalog_url)
-        self.assertEqual(None, v2_data.min_microversion)
-        self.assertEqual(None, v2_data.max_microversion)
+        self.assertIsNone(v2_data.min_microversion)
+        self.assertIsNone(v2_data.max_microversion)
 
     def test_get_versioned_data_volume_project_id_unversioned_first(self):
 
@@ -730,8 +856,8 @@ class CommonIdentityTests(object):
         self.assertEqual(self.TEST_VOLUME.versions['v2'].service.public,
                          v2_data.service_url)
         self.assertEqual(self.TEST_VOLUME.catalog.public, v2_data.catalog_url)
-        self.assertEqual(None, v2_data.min_microversion)
-        self.assertEqual(None, v2_data.max_microversion)
+        self.assertIsNone(v2_data.min_microversion)
+        self.assertIsNone(v2_data.max_microversion)
 
         # Since we fetched from the unversioned endpoint to satisfy the
         # request for v2, we should have all the relevant data cached in the
@@ -748,6 +874,84 @@ class CommonIdentityTests(object):
         self.assertEqual(self.TEST_VOLUME.versions['v3'].service.public,
                          v3_data.service_url)
 
+    def test_trailing_slash_on_computed_endpoint(self):
+
+        disc = fixture.DiscoveryList(v2=False, v3=False)
+
+        # A versioned URL in the Catalog
+        disc.add_nova_microversion(
+            href=self.TEST_VOLUME.versions['v3'].discovery.public,
+            id='v3.0', status='CURRENT',
+            min_version='3.0', version='3.20')
+
+        a = self.create_auth_plugin()
+        s = session.Session(auth=a)
+
+        # endpoint ends in v3, we will construct the unversioned endpoint.
+        # Because the catalog has the versioned endpoint but we constructed
+        # an unversioned endpoint, the url needs to have a trailing /
+        self.requests_mock.get(
+            self.TEST_VOLUME.unversioned.public + '/', json=disc)
+
+        # We're requesting version 2 of volumev3 to make sure we
+        # trigger the logic constructing the unversioned endpoint from the
+        # versioned endpoint in the catalog
+        s.get_endpoint_data(service_type='volumev3',
+                            interface='public',
+                            min_version='2.0',
+                            max_version='2.latest',
+                            project_id=self.project_id)
+
+        self.assertTrue(
+            self.requests_mock.request_history[-1].url.endswith('/'))
+
+    def test_no_trailing_slash_on_catalog_endpoint(self):
+
+        disc = fixture.DiscoveryList(v2=False, v3=False)
+
+        # A versioned URL in the Catalog
+        disc.add_nova_microversion(
+            href=self.TEST_COMPUTE_PUBLIC,
+            id='v2.1', status='CURRENT',
+            min_version='2.1', version='2.38')
+
+        a = self.create_auth_plugin()
+        s = session.Session(auth=a)
+
+        # nova has unversioned endpoint in this catalog. We should not
+        # modify it.
+        self.requests_mock.get(self.TEST_COMPUTE_PUBLIC, json=disc)
+
+        s.get_endpoint_data(service_type='compute',
+                            interface='public',
+                            min_version='2.1',
+                            max_version='2.latest')
+
+        self.assertFalse(
+            self.requests_mock.request_history[-1].url.endswith('/'))
+
+    def test_broken_discovery_endpoint(self):
+
+        # Discovery document with a bogus/broken base URL
+        disc = fixture.DiscoveryList(v2=False, v3=False)
+        disc.add_nova_microversion(
+            href='http://internal.example.com',
+            id='v2.1', status='CURRENT',
+            min_version='2.1', version='2.38')
+
+        a = self.create_auth_plugin()
+        s = session.Session(auth=a)
+
+        self.requests_mock.get(self.TEST_COMPUTE_PUBLIC, json=disc)
+
+        data = s.get_endpoint_data(service_type='compute',
+                                   interface='public',
+                                   min_version='2.1',
+                                   max_version='2.latest')
+
+        # Verify that we get the versioned url based on the catalog url
+        self.assertTrue(data.url, self.TEST_COMPUTE_PUBLIC + '/v2.1')
+
     def test_asking_for_auth_endpoint_ignores_checks(self):
         a = self.create_auth_plugin()
         s = session.Session(auth=a)
@@ -876,8 +1080,9 @@ class CommonIdentityTests(object):
 
         data = sess.get_endpoint_data(service_type='network')
 
-        # Discovery ran and returned the URL
+        # Discovery ran and returned the URL and its version
         self.assertEqual(url, data.url)
+        self.assertEqual((2, 1), data.api_version)
 
         # Run with a project_id to ensure that path is covered
         self.assertEqual(
@@ -1013,6 +1218,8 @@ class CatalogHackTests(utils.TestCase):
     V2_URL = BASE_URL + 'v2.0'
     V3_URL = BASE_URL + 'v3'
 
+    PROJECT_ID = uuid.uuid4().hex
+
     def test_getting_endpoints(self):
         disc = fixture.DiscoveryList(href=self.BASE_URL)
         self.stub_url('GET',
@@ -1070,6 +1277,43 @@ class CatalogHackTests(utils.TestCase):
 
         self.assertEqual(self.V2_URL, endpoint)
 
+    def test_getting_endpoints_project_id_and_trailing_slash_in_disc_url(self):
+        # Test that when requesting a v3 endpoint and having a project in the
+        # session but only the v2 endpoint with a trailing slash in the
+        # catalog, we can still discover the v3 endpoint.
+        disc = fixture.DiscoveryList(href=self.BASE_URL)
+        self.stub_url('GET',
+                      ['/'],
+                      base_url=self.BASE_URL,
+                      json=disc)
+
+        # Create a project-scoped token. This will exercise the flow in the
+        # discovery URL sequence where a project ID exists in the token but
+        # there is no project ID in the URL.
+        token = fixture.V3Token(project_id=self.PROJECT_ID)
+
+        # Add only a v2 endpoint with a trailing slash
+        service = token.add_service(self.IDENTITY)
+        service.add_endpoint('public', self.V2_URL + '/')
+        service.add_endpoint('admin', self.V2_URL + '/')
+
+        # Auth with v3
+        kwargs = {'headers': {'X-Subject-Token': self.TEST_TOKEN}}
+        self.stub_url('POST',
+                      ['auth', 'tokens'],
+                      base_url=self.V3_URL,
+                      json=token, **kwargs)
+        v3_auth = identity.V3Password(self.V3_URL,
+                                      username=uuid.uuid4().hex,
+                                      password=uuid.uuid4().hex)
+        sess = session.Session(auth=v3_auth)
+
+        # Try to get a v3 endpoint
+        endpoint = sess.get_endpoint(service_type=self.IDENTITY,
+                                     interface='public',
+                                     version=(3, 0))
+        self.assertEqual(self.V3_URL, endpoint)
+
     def test_returns_original_skipping_discovery(self):
         token = fixture.V2Token()
         service = token.add_service(self.IDENTITY)
@@ -1157,6 +1401,7 @@ class CatalogHackTests(utils.TestCase):
         self.assertEqual(self.OTHER_URL, data.url)
         self.assertEqual((2, 1), data.min_microversion)
         self.assertEqual((2, 35), data.max_microversion)
+        self.assertEqual((2, 1), data.api_version)
 
     def test_forcing_discovery(self):
         v2_disc = fixture.V2Discovery(self.V2_URL)
@@ -1204,6 +1449,7 @@ class CatalogHackTests(utils.TestCase):
 
         # got v2 url
         self.assertEqual(self.V2_URL, data.url)
+        self.assertEqual((2, 0), data.api_version)
 
     def test_forcing_discovery_list_returns_url(self):
         common_disc = fixture.DiscoveryList(href=self.BASE_URL)
@@ -1248,6 +1494,7 @@ class CatalogHackTests(utils.TestCase):
 
         # got v2 url
         self.assertEqual(self.V2_URL, data.url)
+        self.assertEqual((2, 0), data.api_version)
 
     def test_latest_version_gets_latest_version(self):
         common_disc = fixture.DiscoveryList(href=self.BASE_URL)
@@ -1371,7 +1618,7 @@ class CatalogHackTests(utils.TestCase):
         self.assertFalse(common_m.called)
 
         # And get no url
-        self.assertEqual(None, endpoint)
+        self.assertIsNone(endpoint)
 
         v2_m, common_m = stub_urls()
         endpoint = sess.get_endpoint(service_type=self.IDENTITY,
@@ -1433,6 +1680,7 @@ class CatalogHackTests(utils.TestCase):
         self.assertEqual(self.OTHER_URL, data.url)
         self.assertEqual((2, 1), data.min_microversion)
         self.assertEqual((2, 35), data.max_microversion)
+        self.assertEqual((2, 1), data.api_version)
 
     def test_get_endpoint_data_compute(self):
         common_disc = fixture.DiscoveryList(v2=False, v3=False)
@@ -1474,6 +1722,7 @@ class CatalogHackTests(utils.TestCase):
         self.assertEqual(self.OTHER_URL, data.url)
         self.assertEqual((2, 1), data.min_microversion)
         self.assertEqual((2, 35), data.max_microversion)
+        self.assertEqual((2, 1), data.api_version)
 
     def test_getting_endpoints_on_auth_interface(self):
         disc = fixture.DiscoveryList(href=self.BASE_URL)
@@ -1679,3 +1928,27 @@ class GenericAuthPluginTests(utils.TestCase):
                               endpoint_filter=self.ENDPOINT_FILTER)
 
         self.assertIn(name, str(e))
+
+
+class DiscoveryFailures(utils.TestCase):
+    TEST_ROOT_URL = 'http://127.0.0.1:5000/'
+
+    def test_connection_error(self):
+        self.requests_mock.get(self.TEST_ROOT_URL,
+                               exc=exceptions.ConnectionError)
+        sess = session.Session()
+        p = identity.generic.password.Password(self.TEST_ROOT_URL)
+        self.assertRaises(exceptions.DiscoveryFailure, p.get_auth_ref, sess)
+
+    def test_client_exception(self):
+        self.requests_mock.get(self.TEST_ROOT_URL,
+                               exc=exceptions.ClientException)
+        sess = session.Session()
+        p = identity.generic.password.Password(self.TEST_ROOT_URL)
+        self.assertRaises(exceptions.ClientException, p.get_auth_ref, sess)
+
+    def test_ssl_error(self):
+        self.requests_mock.get(self.TEST_ROOT_URL, exc=exceptions.SSLError)
+        sess = session.Session()
+        p = identity.generic.password.Password(self.TEST_ROOT_URL)
+        self.assertRaises(exceptions.DiscoveryFailure, p.get_auth_ref, sess)
