@@ -23,6 +23,7 @@ import six
 from testtools import matchers
 
 from keystoneauth1 import adapter
+from keystoneauth1 import discover
 from keystoneauth1 import exceptions
 from keystoneauth1 import plugin
 from keystoneauth1 import session as client_session
@@ -105,6 +106,112 @@ class SessionTests(utils.TestCase):
         self.assertEqual(resp.text, 'response')
         self.assertRequestBodyIs(json={'hello': 'world'})
 
+    def test_set_microversion_headers(self):
+
+        # String microversion, specified service type
+        headers = {}
+        client_session.Session._set_microversion_headers(
+            headers, '2.30', 'compute', None)
+        self.assertEqual(headers['OpenStack-API-Version'], 'compute 2.30')
+        self.assertEqual(headers['X-OpenStack-Nova-API-Version'], '2.30')
+        self.assertEqual(len(headers.keys()), 2)
+
+        # Tuple microversion, service type via endpoint_filter
+        headers = {}
+        client_session.Session._set_microversion_headers(
+            headers, (2, 30), None, {'service_type': 'compute'})
+        self.assertEqual(headers['OpenStack-API-Version'], 'compute 2.30')
+        self.assertEqual(headers['X-OpenStack-Nova-API-Version'], '2.30')
+        self.assertEqual(len(headers.keys()), 2)
+
+        # 'latest' (string) microversion
+        headers = {}
+        client_session.Session._set_microversion_headers(
+            headers, 'latest', 'compute', None)
+        self.assertEqual(headers['OpenStack-API-Version'], 'compute latest')
+        self.assertEqual(headers['X-OpenStack-Nova-API-Version'], 'latest')
+        self.assertEqual(len(headers.keys()), 2)
+
+        # LATEST (tuple) microversion
+        headers = {}
+        client_session.Session._set_microversion_headers(
+            headers, (discover.LATEST, discover.LATEST), 'compute', None)
+        self.assertEqual(headers['OpenStack-API-Version'], 'compute latest')
+        self.assertEqual(headers['X-OpenStack-Nova-API-Version'], 'latest')
+        self.assertEqual(len(headers.keys()), 2)
+
+        # ironic microversion, specified service type
+        headers = {}
+        client_session.Session._set_microversion_headers(
+            headers, '2.30', 'baremetal', None)
+        self.assertEqual(headers['OpenStack-API-Version'], 'baremetal 2.30')
+        self.assertEqual(headers['X-OpenStack-Ironic-API-Version'], '2.30')
+        self.assertEqual(len(headers.keys()), 2)
+
+        # volumev2 service-type - volume microversion
+        headers = {}
+        client_session.Session._set_microversion_headers(
+            headers, (2, 30), None, {'service_type': 'volumev2'})
+        self.assertEqual(headers['OpenStack-API-Version'], 'volume 2.30')
+        self.assertEqual(len(headers.keys()), 1)
+
+        # block-storage service-type - volume microversion
+        headers = {}
+        client_session.Session._set_microversion_headers(
+            headers, (2, 30), None, {'service_type': 'block-storage'})
+        self.assertEqual(headers['OpenStack-API-Version'], 'volume 2.30')
+        self.assertEqual(len(headers.keys()), 1)
+
+        # Headers already exist - no change
+        headers = {
+            'OpenStack-API-Version': 'compute 2.30',
+            'X-OpenStack-Nova-API-Version': '2.30',
+        }
+        client_session.Session._set_microversion_headers(
+            headers, (2, 31), None, {'service_type': 'volume'})
+        self.assertEqual(headers['OpenStack-API-Version'], 'compute 2.30')
+        self.assertEqual(headers['X-OpenStack-Nova-API-Version'], '2.30')
+
+        # Can't specify a 'M.latest' microversion
+        self.assertRaises(TypeError,
+                          client_session.Session._set_microversion_headers,
+                          {}, '2.latest', 'service_type', None)
+        self.assertRaises(TypeError,
+                          client_session.Session._set_microversion_headers,
+                          {}, (2, discover.LATEST), 'service_type', None)
+
+        # Normalization error
+        self.assertRaises(TypeError,
+                          client_session.Session._set_microversion_headers,
+                          {}, 'bogus', 'service_type', None)
+
+        # No service type in param or endpoint filter
+        self.assertRaises(TypeError,
+                          client_session.Session._set_microversion_headers,
+                          {}, (2, 30), None, None)
+        self.assertRaises(TypeError,
+                          client_session.Session._set_microversion_headers,
+                          {}, (2, 30), None, {'no_service_type': 'here'})
+
+    def test_microversion(self):
+        # microversion not specified
+        session = client_session.Session()
+        self.stub_url('GET', text='response')
+        resp = session.get(self.TEST_URL)
+
+        self.assertTrue(resp.ok)
+        self.assertRequestNotInHeader('OpenStack-API-Version')
+
+        session = client_session.Session()
+        self.stub_url('GET', text='response')
+        resp = session.get(self.TEST_URL, microversion='2.30',
+                           microversion_service_type='compute',
+                           endpoint_filter={'endpoint': 'filter'})
+
+        self.assertTrue(resp.ok)
+        self.assertRequestHeaderEqual('OpenStack-API-Version', 'compute 2.30')
+        self.assertRequestHeaderEqual('X-OpenStack-Nova-API-Version', '2.30')
+
     def test_user_agent(self):
         session = client_session.Session()
         self.stub_url('GET', text='response')
@@ -219,6 +326,33 @@ class SessionTests(utils.TestCase):
             self.assertEqual(v, resp.headers[k])
             self.assertNotIn(v, self.logger.output)
 
+    def test_session_debug_output_logs_openstack_request_id(self):
+        """Test x-openstack-request-id is logged in debug logs."""
+        def get_response(log=True):
+            session = client_session.Session(verify=False)
+            endpoint_filter = {'service_name': 'Identity'}
+            headers = {'X-OpenStack-Request-Id': 'req-1234'}
+            body = 'BODYRESPONSE'
+            data = 'BODYDATA'
+            all_headers = dict(itertools.chain(headers.items()))
+            self.stub_url('POST', text=body, headers=all_headers)
+            resp = session.post(self.TEST_URL, endpoint_filter=endpoint_filter,
+                                headers=all_headers, data=data, log=log)
+            return resp
+
+        # if log is disabled then request-id is not logged in debug logs
+        resp = get_response(log=False)
+        self.assertEqual(resp.status_code, 200)
+
+        expected_log = ('POST call to Identity for %s used request '
+                        'id req-1234' % self.TEST_URL)
+        self.assertNotIn(expected_log, self.logger.output)
+
+        # if log is enabled then request-id is logged in debug logs
+        resp = get_response()
+        self.assertEqual(resp.status_code, 200)
+        self.assertIn(expected_log, self.logger.output)
+
     def test_logs_failed_output(self):
         """Test that output is logged even for failed requests."""
         session = client_session.Session()
@@ -265,6 +399,16 @@ class SessionTests(utils.TestCase):
         self.assertIn(body, self.logger.output)
         self.assertNotIn(OMITTED_BODY % 'application/json', self.logger.output)
 
+        # Content-Type is set to application/json; charset=UTF-8
+        body = json.dumps({'token': {'id': '...'}})
+        self.stub_url(
+            'POST', text=body,
+            headers={'Content-Type': 'application/json; charset=UTF-8'})
+        session.post(self.TEST_URL)
+        self.assertIn(body, self.logger.output)
+        self.assertNotIn(OMITTED_BODY % 'application/json; charset=UTF-8',
+                         self.logger.output)
+
     def test_logging_cacerts(self):
         path_to_certs = '/path/to/certs'
         session = client_session.Session(verify=path_to_certs)
@@ -812,6 +956,60 @@ class SessionAuthTests(utils.TestCase):
         self.assertNotIn(list(response.keys())[0], self.logger.output)
         self.assertNotIn(list(response.values())[0], self.logger.output)
 
+    def test_split_loggers(self):
+
+        def get_logger_io(name):
+            logger_name = 'keystoneauth.session.{name}'.format(name=name)
+            logger = logging.getLogger(logger_name)
+            logger.setLevel(logging.DEBUG)
+
+            io = six.StringIO()
+            handler = logging.StreamHandler(io)
+            logger.addHandler(handler)
+            return io
+
+        io = {}
+        for name in ('request', 'body', 'response', 'request-id'):
+            io[name] = get_logger_io(name)
+
+        auth = AuthPlugin()
+        sess = client_session.Session(auth=auth, split_loggers=True)
+        response_key = uuid.uuid4().hex
+        response_val = uuid.uuid4().hex
+        response = {response_key: response_val}
+        request_id = uuid.uuid4().hex
+
+        self.stub_url(
+            'GET',
+            json=response,
+            headers={
+                'Content-Type': 'application/json',
+                'X-OpenStack-Request-ID': request_id,
+            })
+
+        resp = sess.get(self.TEST_URL)
+
+        self.assertEqual(response, resp.json())
+
+        request_output = io['request'].getvalue().strip()
+        response_output = io['response'].getvalue().strip()
+        body_output = io['body'].getvalue().strip()
+        id_output = io['request-id'].getvalue().strip()
+
+        self.assertIn('curl -g -i -X GET {url}'.format(url=self.TEST_URL),
+                      request_output)
+        self.assertEqual('[200] Content-Type: application/json '
+                         'X-OpenStack-Request-ID: '
+                         '{id}'.format(id=request_id), response_output)
+        self.assertEqual(
+            'GET call to {url} used request id {id}'.format(
+                url=self.TEST_URL, id=request_id),
+            id_output)
+        self.assertEqual(
+            '{{"{key}": "{val}"}}'.format(
+                key=response_key, val=response_val),
+            body_output)
+
 
 class AdapterTest(utils.TestCase):
 
@@ -863,6 +1061,29 @@ class AdapterTest(utils.TestCase):
         self.assertTrue(adpt.auth.get_token_called)
         self.assertRequestHeaderEqual('User-Agent', self.USER_AGENT)
 
+    def test_setting_global_id_on_request(self):
+        global_id = "req-%s" % uuid.uuid4()
+        response = uuid.uuid4().hex
+        self.stub_url('GET', text=response)
+        adpt = adapter.Adapter(client_session.Session(),
+                               auth=CalledAuthPlugin(),
+                               service_type=self.SERVICE_TYPE,
+                               service_name=self.SERVICE_NAME,
+                               interface=self.INTERFACE,
+                               region_name=self.REGION_NAME,
+                               user_agent=self.USER_AGENT,
+                               version=self.VERSION,
+                               allow=self.ALLOW,
+                               global_request_id=global_id)
+        resp = adpt.get('/')
+        self.assertEqual(resp.text, response)
+
+        self._verify_endpoint_called(adpt)
+        self.assertEqual(self.ALLOW,
+                         adpt.auth.endpoint_arguments['allow'])
+        self.assertTrue(adpt.auth.get_token_called)
+        self.assertRequestHeaderEqual('X-OpenStack-Request-ID', global_id)
+
     def test_setting_variables_on_get_endpoint(self):
         adpt = self._create_loaded_adapter()
         url = adpt.get_endpoint()
@@ -1140,7 +1361,7 @@ class AdapterTest(utils.TestCase):
 
         adap.get(url)
 
-        agent = 'ksatest/1.2.3 one/1.1.1 two/2.2.2 testclient/4.5.6'
+        agent = 'ksatest/1.2.3 testclient/4.5.6 one/1.1.1 two/2.2.2'
         self.assertEqual(agent + ' ' + client_session.DEFAULT_USER_AGENT,
                          self.requests_mock.last_request.headers['User-Agent'])
 
@@ -1219,6 +1440,29 @@ class AdapterTest(utils.TestCase):
         last_token = self.requests_mock.last_request.headers['X-Auth-Token']
         self.assertEqual(token, last_token)
 
+    def test_default_microversion(self):
+        sess = client_session.Session()
+        url = 'http://url'
+
+        def validate(adap_kwargs, get_kwargs, exp_kwargs):
+            with mock.patch.object(sess, 'request') as m:
+                adapter.Adapter(sess, **adap_kwargs).get(url, **get_kwargs)
+                m.assert_called_once_with(url, 'GET', endpoint_filter={},
+                                          **exp_kwargs)
+
+        # No default_microversion in Adapter, no microversion in get()
+        validate({}, {}, {})
+
+        # default_microversion in Adapter, no microversion in get()
+        validate({'default_microversion': '1.2'}, {}, {'microversion': '1.2'})
+
+        # No default_microversion in Adapter, microversion specified in get()
+        validate({}, {'microversion': '1.2'}, {'microversion': '1.2'})
+
+        # microversion in get() overrides default_microversion in Adapter
+        validate({'default_microversion': '1.2'}, {'microversion': '1.5'},
+                 {'microversion': '1.5'})
+
 
 class TCPKeepAliveAdapterTest(utils.TestCase):
 
